
//******************************************************************************
// RCF - Remote Call Framework
//
// Copyright (c) 2005 - 2012, Delta V Software. All rights reserved.
// http://www.deltavsoft.com
//
// RCF is distributed under dual licenses - closed source or GPL.
// Consult your particular license for conditions of use.
//
// If you have not purchased a commercial license, you are using RCF 
// under GPL terms.
//
// Version: 2.0
// Contact: support <at> deltavsoft.com 
//
//******************************************************************************

#ifndef INCLUDE_RCF_AMITHREADPOOL_HPP
#define INCLUDE_RCF_AMITHREADPOOL_HPP

#include <memory>

#include <boost/bind.hpp>
#include <boost/function.hpp>
#include <boost/scoped_ptr.hpp>

#include <RCF/Asio.hpp>
#include <RCF/Export.hpp>
#include <RCF/ThreadLibrary.hpp>
#include <RCF/ThreadPool.hpp>
#include <RCF/Tools.hpp>

namespace RCF {

    class RCF_EXPORT AmiNotification
    {
    public:

        typedef boost::function0<void> Cb;

        // Need mutexPtr so that the mutex doesn't die before the lock.
        void set(Cb cb, LockPtr lockPtr, MutexPtr mutexPtr);
        void run();
        void clear();

    private:
        Cb                  mCb;
        MutexPtr   mMutexPtr;
        LockPtr    mLockPtr;
    };

} // namespace RCF

namespace RCF {

    class OverlappedAmi;
    typedef boost::shared_ptr<OverlappedAmi> OverlappedAmiPtr;

    class ConnectionOrientedClientTransport;

    class OverlappedAmi : 
        public boost::enable_shared_from_this<OverlappedAmi>
    {
    public:

        enum AsyncOpType
        {
            None,
            Wait,
            Connect,
            Write,
            Read
        };

        OverlappedAmi(ConnectionOrientedClientTransport *pTcpClientTransport) : 
            mpTransport(pTcpClientTransport),
            mIndex(0),
            mOpType(None)
        {
        }

        ~OverlappedAmi()
        {
        }

        void onCompletion(
            std::size_t index,
            const AsioErrorCode & ec, 
            std::size_t bytesTransferred);

        void onTimerExpired(
            std::size_t index,
            const AsioErrorCode & ec);

        RecursiveMutex                      mMutex;
        ConnectionOrientedClientTransport * mpTransport;
        std::size_t                         mIndex;
        AsyncOpType                         mOpType;
    };

    class AmiIoHandler
    {
    public:

        AmiIoHandler(OverlappedAmiPtr overlappedPtr);

        AmiIoHandler(OverlappedAmiPtr overlappedPtr, const AsioErrorCode & ec);

        void operator()();

        void operator()(
            const AsioErrorCode & ec);

        void operator()(
            const AsioErrorCode & ec, 
            std::size_t bytesTransferred);

        OverlappedAmiPtr mOverlappedPtr;
        std::size_t mIndex;

        AsioErrorCode mEc;
    };

    class AmiTimerHandler
    {
    public:

        AmiTimerHandler(OverlappedAmiPtr overlappedPtr);

        void operator()(const AsioErrorCode & ec);

        OverlappedAmiPtr mOverlappedPtr;
        std::size_t mIndex;
    };

    void *  asio_handler_allocate(std::size_t size, AmiIoHandler * pHandler);
    void    asio_handler_deallocate(void * pointer, std::size_t size, AmiIoHandler * pHandler);
    void *  asio_handler_allocate(std::size_t size, AmiTimerHandler * pHandler);
    void    asio_handler_deallocate(void * pointer, std::size_t size, AmiTimerHandler * pHandler);

    class RCF_EXPORT AmiThreadPool
    {
    public:
        AmiThreadPool();
        ~AmiThreadPool();

        void start();
        void stop();

        AsioIoService & getIoService();

        static void setEnableMultithreadedBlockingConnects(bool enable = true);
        static bool getEnableMultithreadedBlockingConnects();

    private:
        ThreadPool mThreadPool;
    };

    RCF_EXPORT AmiThreadPool & getAmiThreadPool();

    class I_RcfClient;

    // Deprecated.
    class RCF_EXPORT AsyncTimer
    {
    public:
        AsyncTimer();
        ~AsyncTimer();
        void set(boost::function0<void> onTimer, boost::uint32_t timeoutMs);
        void cancel();

    private:
        boost::shared_ptr<I_RcfClient> mDummyPtr;

    };

} // namespace RCF

#endif // ! INCLUDE_RCF_AMITHREADPOOL_HPP
