
//******************************************************************************
// RCF - Remote Call Framework
//
// Copyright (c) 2005 - 2012, Delta V Software. All rights reserved.
// http://www.deltavsoft.com
//
// RCF is distributed under dual licenses - closed source or GPL.
// Consult your particular license for conditions of use.
//
// If you have not purchased a commercial license, you are using RCF 
// under GPL terms.
//
// Version: 2.0
// Contact: support <at> deltavsoft.com 
//
//******************************************************************************

#include <RCF/BsdClientTransport.hpp>

#include <RCF/AmiThreadPool.hpp>
#include <RCF/ClientStub.hpp>
#include <RCF/Exception.hpp>
#include <RCF/RcfServer.hpp>
#include <RCF/ThreadLocalData.hpp>
#include <RCF/TimedBsdSockets.hpp>

#include <RCF/AsioServerTransport.hpp>

namespace RCF {

    BsdClientTransport::BsdClientTransport() :
        mFd(-1),
        mpIoService(NULL)
    {}

    BsdClientTransport::BsdClientTransport(TcpSocketPtr socketPtr) :
        mFd(-1),
        mSocketPtr(socketPtr),
        mpIoService(& socketPtr->get_io_service())
    {
        mClosed = false;
        mAsioTimerPtr.reset( new AsioDeadlineTimer( getAmiThreadPool().getIoService() ));
    }

    BsdClientTransport::BsdClientTransport(const BsdClientTransport & rhs) :
        ConnectionOrientedClientTransport(rhs),
        mFd(-1),
        mpIoService(NULL)
    {}

    BsdClientTransport::~BsdClientTransport()
    {
        RCF_DTOR_BEGIN
            close();
        RCF_DTOR_END
    }

#ifdef BOOST_WINDOWS

    // return -2 for timeout, -1 for error, 0 for ready
    int pollSocketWithProgressMwfmo(
        const ClientProgressPtr &clientProgressPtr,
        ClientProgress::Activity activity,
        unsigned int endTimeMs,
        int fd,
        int &err,
        bool bRead)
    {
        RCF_UNUSED_VARIABLE(err);
        RCF_UNUSED_VARIABLE(activity);

        ClientStub & clientStub = *getTlsClientStubPtr();

        int uiMessageFilter = clientProgressPtr->mUiMessageFilter;

        HANDLE readEvent = WSACreateEvent();
        using namespace boost::multi_index::detail;
        scope_guard WSACloseEventGuard = make_guard(WSACloseEvent, readEvent);
        RCF_UNUSED_VARIABLE(WSACloseEventGuard);

        int nRet = WSAEventSelect(fd, readEvent, bRead ? FD_READ : FD_WRITE);
        RCF_ASSERT_EQ(nRet , 0);
        HANDLE handles[] = { readEvent };

        while (true)
        {
            unsigned int timeoutMs = generateTimeoutMs(endTimeMs);
            timeoutMs = clientStub.generatePollingTimeout(timeoutMs);

            DWORD dwRet = MsgWaitForMultipleObjects(
                1, 
                handles, 
                0, 
                timeoutMs, 
                uiMessageFilter);

            if (dwRet == WAIT_TIMEOUT)
            {
                clientStub.onPollingTimeout();

                if (generateTimeoutMs(endTimeMs) == 0)
                {
                    return -2;
                }
            }
            else if (dwRet == WAIT_OBJECT_0)
            {
                // File descriptor is ready to be read.
                return 0;
            }
            else if (dwRet == WAIT_OBJECT_0 + 1)
            {
                clientStub.onUiMessage();                
            }
        }
    }

#endif

    PollingFunctor::PollingFunctor(
        ClientProgressPtr clientProgressPtr,
        ClientProgress::Activity activity,
        unsigned int endTimeMs) :
            mClientProgressPtr(clientProgressPtr),
            mActivity(activity),
            mEndTimeMs(endTimeMs)
    {}

#ifdef BOOST_WINDOWS

    // On Windows, the user may have requested callbacks on UI messages, in 
    // which case we'll need to use MsgWaitForMultipleObjects() rather than
    // plain old select().

    int PollingFunctor::operator()(int fd, int &err, bool bRead)
    {
        if (
            mClientProgressPtr.get() &&
            mClientProgressPtr->mTriggerMask & ClientProgress::UiMessage)
        {
            return pollSocketWithProgressMwfmo(
                mClientProgressPtr,
                mActivity,
                mEndTimeMs,
                fd,
                err,
                bRead);
        }
        else
        {
            return pollSocket(
                mEndTimeMs,
                fd,
                err,
                bRead);
        }
    }

#else

    int PollingFunctor::operator()(int fd, int &err, bool bRead)
    {
        return pollSocket(
            mEndTimeMs,
            fd,
            err,
            bRead);
    }

#endif

    std::size_t BsdClientTransport::implRead(
        const ByteBuffer &byteBuffer,
        std::size_t bytesRequested)
    {
        std::size_t bytesToRead = RCF_MIN(bytesRequested, byteBuffer.getLength());

        PollingFunctor pollingFunctor(
            mClientProgressPtr,
            ClientProgress::Receive,
            mEndTimeMs);

        RCF_LOG_4()(byteBuffer.getLength())(bytesToRead) << "BsdClientTransport - initiating read from socket.";

        int fd = getNativeHandle();

        int err = 0;
        int ret = RCF::timedRecv(
            *this,
            pollingFunctor,
            err,
            fd,
            byteBuffer,
            bytesToRead,
            0);

        switch (ret)
        {
        case -2:
            {
                Exception e(_RcfError_ClientReadTimeout());
                RCF_THROW(e);
            }
            break;

        case -1:
            {
                Exception e(
                    _RcfError_ClientReadFail(),
                    err,
                    RcfSubsystem_Os);
            
                RCF_THROW(e);
            }
            break;

        case  0:
            {
                Exception e(_RcfError_PeerDisconnect());
                RCF_THROW(e);
            }
            break;

        default:
            
            RCF_ASSERT(
                0 < ret && ret <= static_cast<int>(bytesRequested))
                (ret)(bytesRequested);
        }

        return ret;
    }

    std::size_t BsdClientTransport::implReadAsync(
        const ByteBuffer &byteBuffer,
        std::size_t bytesRequested)
    {
        RecursiveLock lock(mOverlappedPtr->mMutex);

        mOverlappedPtr->mOpType = OverlappedAmi::Read;

        mSocketPtr->async_read_some(
            ASIO_NS::buffer( byteBuffer.getPtr(), bytesRequested),
            AmiIoHandler(mOverlappedPtr));

        boost::uint32_t nowMs = getCurrentTimeMs();
        boost::uint32_t timeoutMs = mEndTimeMs - nowMs;
        mAsioTimerPtr->expires_from_now( boost::posix_time::milliseconds(timeoutMs) );

        mAsioTimerPtr->async_wait( AmiTimerHandler(mOverlappedPtr) );

        return 0;
    }

    std::size_t BsdClientTransport::implWrite(
        const std::vector<ByteBuffer> &byteBuffers)
    {
        PollingFunctor pollingFunctor(
            mClientProgressPtr,
            ClientProgress::Send,
            mEndTimeMs);

        int err = 0;

        RCF_LOG_4()(lengthByteBuffers(byteBuffers)) 
            << "BsdClientTransport - initiating send on socket.";

        int fd = getNativeHandle();

        int ret = RCF::timedSend(
            pollingFunctor,
            err,
            fd,
            byteBuffers,
            getMaxSendSize(),
            0);

        switch (ret)
        {
        case -2:
            {
                Exception e(_RcfError_ClientWriteTimeout());
                RCF_THROW(e);
            }
            break;

        case -1:
            {
                Exception e(
                    _RcfError_ClientWriteFail(),
                    err,
                    RcfSubsystem_Os);
            
                RCF_THROW(e);
            }
            break;

        case 0:
            {
                Exception e(_RcfError_PeerDisconnect());            
                RCF_THROW(e);
            }
            break;

        default:
            
            RCF_ASSERT(
                0 < ret && ret <= static_cast<int>(lengthByteBuffers(byteBuffers)))
                (ret)(lengthByteBuffers(byteBuffers));            

            onTimedSendCompleted(ret, 0);
        }

        return ret;
    }

    std::size_t BsdClientTransport::implWriteAsync(
        const std::vector<ByteBuffer> &byteBuffers)
    {
        if (!mAsioBuffersPtr)
        {
            mAsioBuffersPtr.reset( new AsioBuffers() );
        }

        mAsioBuffersPtr->mVecPtr->resize(0);

        for (std::size_t i=0; i<byteBuffers.size(); ++i)
        {
            const ByteBuffer & buffer = byteBuffers[i];
            mAsioBuffersPtr->mVecPtr->push_back( AsioConstBuffer(
                buffer.getPtr(), 
                buffer.getLength()));
        }

        RecursiveLock lock(mOverlappedPtr->mMutex);

        mOverlappedPtr->mOpType = OverlappedAmi::Write;

        ASIO_NS::async_write(
            *mSocketPtr, 
            *mAsioBuffersPtr, 
            AmiIoHandler(mOverlappedPtr));


        boost::uint32_t nowMs = getCurrentTimeMs();
        boost::uint32_t timeoutMs = mEndTimeMs - nowMs;
        mAsioTimerPtr->expires_from_now( boost::posix_time::milliseconds(timeoutMs) );

        mAsioTimerPtr->async_wait( AmiTimerHandler(mOverlappedPtr) );

        return 0;
    }

    bool BsdClientTransport::isConnected()
    {
        int fd = getNativeHandle();
        return isFdConnected(fd);
    }

    BsdClientTransport::TcpSocketPtr BsdClientTransport::releaseSocket()
    {
        RCF_ASSERT( mFd == -1 );
        RCF_ASSERT( mSocketPtr );

        TcpSocketPtr socketPtr = mSocketPtr;
        mSocketPtr.reset();
        return socketPtr;
    }

    int BsdClientTransport::getNativeHandle() const
    {
        if (mSocketPtr)
        {
            return static_cast<int>(mSocketPtr->native());
        }
        else
        {
            return mFd;
        }
    }
    
} // namespace RCF
